#include "gdal_pam.h"
#include "gdal_proxy.h"
#include "ogr_spatialref.h"
#include "ogr_geometry.h"
#include "stdio.h"

CPL_CVSID("$Id$");

CPL_C_START
void	GDALRegister_PSP(void);
CPL_C_END

void cp2utf( char* str, char* res );

/************************************************************************/
/* ==================================================================== */
/*                         PSPWrapperRasterBand                         */
/* ==================================================================== */
/************************************************************************/
class PSPWrapperRasterBand : public GDALProxyRasterBand
{
  GDALRasterBand* poBaseBand;

  protected:
    virtual GDALRasterBand* RefUnderlyingRasterBand() { return poBaseBand; }

  public:
    PSPWrapperRasterBand( GDALRasterBand* poBaseBand )
        {
            this->poBaseBand = poBaseBand;
            eDataType = poBaseBand->GetRasterDataType();
            poBaseBand->GetBlockSize(&nBlockXSize, &nBlockYSize);
        }
    ~PSPWrapperRasterBand() {}
};

class PSPDataset : public GDALDataset
{

    GDALDataset *poImageDS;

    double      adfGeoTransform[6];
    CPLString   osImgFilename;

  public:
        PSPDataset();
        ~PSPDataset();

        const char* GetProjectionRef();
        CPLErr      GetGeoTransform(double * padfTransform);

        static GDALDataset *Open( GDALOpenInfo * poOpenInfo);
        static int Identify( GDALOpenInfo *poOpenInfo );
};



    ///************************************************************************/
    ///*                              Identify()                              */
    ///************************************************************************/

    int PSPDataset::Identify( GDALOpenInfo *poOpenInfo )

    {
        if( poOpenInfo->nHeaderBytes < 200 || !EQUAL(CPLGetExtension(poOpenInfo->pszFilename),"PSP") )
            return FALSE;

        return TRUE;
    }


    GDALDataset *PSPDataset::Open(GDALOpenInfo * poOpenInfo)
    {

        if( !Identify( poOpenInfo ) )
            return NULL;

        /* -------------------------------------------------------------------- */
        /*      Confirm the requested access is supported.                      */
        /* -------------------------------------------------------------------- */
        if( poOpenInfo->eAccess == GA_Update )
        {
            CPLError( CE_Failure, CPLE_NotSupported,
                      "The PSP driver does not support update access to existing"
                      " datasets.\n" );
            return NULL;
        }

        /* -------------------------------------------------------------------- */
        /*      Create a corresponding GDALDataset.                             */
        /* -------------------------------------------------------------------- */

        PSPDataset *poDS;

        poDS = new PSPDataset();

//        /* -------------------------------------------------------------------- */
//        /*      Try to load and parse the .PSP file.                            */
//        /* -------------------------------------------------------------------- */

        CPLString realfilename;
//        char * realfilename;



        char	**papszLines;
        int		iLine, nLines=0;

        papszLines = CSLLoad2( poOpenInfo->pszFilename, 1000, 200, NULL );

        if ( !papszLines )
            return FALSE;

        nLines = CSLCount( papszLines );

        double fileParam [5] = {0,0,0,0,0};

        if (nLines <= 12)
            return FALSE;


        for ( iLine = 0; iLine < nLines; iLine++ )
        {
            char    **papszTok = NULL;

            if (iLine == 0)
            {
                papszTok = CSLTokenizeString2( papszLines[iLine], "[", CSLT_ALLOWEMPTYTOKENS | CSLT_STRIPLEADSPACES | CSLT_STRIPENDSPACES );
                papszTok = CSLTokenizeString2( papszTok[1], "]", CSLT_ALLOWEMPTYTOKENS | CSLT_STRIPLEADSPACES | CSLT_STRIPENDSPACES );

                realfilename = papszTok[0];
                continue;
            }
            else if (iLine == 1)
            {
                papszTok = CSLTokenizeString2( papszLines[iLine], "=", CSLT_ALLOWEMPTYTOKENS | CSLT_STRIPLEADSPACES | CSLT_STRIPENDSPACES );
            }
            else
            {
                papszTok = CSLTokenizeString2( papszLines[iLine], "=", CSLT_ALLOWEMPTYTOKENS | CSLT_STRIPLEADSPACES | CSLT_STRIPENDSPACES );

                if ( CSLCount(papszTok) < 2 )
                {
                    CSLDestroy(papszTok);
                    continue;
                }

                char * res = new char [1000];
                cp2utf(papszTok[0], res);

                if (!strcmp(res, "МИНХ"))
                {
                    fileParam[0] = CPLAtofM(papszTok[1]);
                }
                else if (!strcmp(res, "МИНУ"))
                {
                    fileParam[1] = CPLAtofM(papszTok[1]);
                }
                else if (!strcmp(res, "МАХХ"))
                {
                    fileParam[2] = CPLAtofM(papszTok[1]);
                }
                else if (!strcmp(res, "МАХУ"))
                {
                    fileParam[3] = CPLAtofM(papszTok[1]);
                }
                else if (!strcmp(res, "ЦРМ"))
                {
                    fileParam[4] = CPLAtofM(papszTok[1]);
                }

                CPLFree ( res );
            }


            CSLDestroy( papszTok );
        }

        CSLDestroy( papszLines );

        poDS->osImgFilename = realfilename;
        VSIStatBufL sStat;
        if (VSIStatL(poDS->osImgFilename, &sStat) != 0)
        {
            CPLString osPath = CPLGetPath(poOpenInfo->pszFilename);
            if (CPLIsFilenameRelative(poDS->osImgFilename))
            {
                poDS->osImgFilename = CPLFormFilename(osPath, poDS->osImgFilename, NULL);
                poDS->osImgFilename.append(".bmp");
            }

        }

        GDALAllRegister();
        poDS->poImageDS = (GDALDataset *) GDALOpen(poDS->osImgFilename, GA_ReadOnly ); //////////////// не открывает файл ЖПЕГ
        if( poDS->poImageDS == NULL || poDS->poImageDS->GetRasterCount() == 0)
        {

            delete poDS;
            return NULL;
        }


        poDS->adfGeoTransform[0] = fileParam [1];
        poDS->adfGeoTransform[3] = fileParam [2];

        poDS->adfGeoTransform[1] = ( fileParam [3] - fileParam [1] ) / poDS->poImageDS->GetRasterXSize();
        poDS->adfGeoTransform[5] = - ( fileParam [2] - fileParam [0] ) / poDS->poImageDS->GetRasterYSize();



        int iBand;

        poDS->nRasterXSize = poDS->poImageDS->GetRasterXSize();
        poDS->nRasterYSize = poDS->poImageDS->GetRasterYSize();

        if (!GDALCheckDatasetDimensions(poDS->nRasterXSize, poDS->nRasterYSize))
        {
            GDALClose( poDS->poImageDS );
            delete poDS;
            return NULL;
        }

        for( iBand = 1; iBand <= poDS->poImageDS->GetRasterCount(); iBand++ )
            poDS->SetBand( iBand,
                           new PSPWrapperRasterBand( poDS->poImageDS->GetRasterBand( iBand )) );

        return( poDS );
    }


/************************************************************************/
/*                          GDALRegister_PSP()                          */
/************************************************************************/



    void GDALRegister_PSP()
    {
        GDALDriver	*poDriver;

        if (! GDAL_CHECK_VERSION("PSP"))
            return;

        if( GDALGetDriverByName( "PSP" ) == NULL )
        {
            poDriver = new GDALDriver();

            poDriver->SetDescription( "PSP" );

            poDriver->pfnOpen = PSPDataset::Open;
            poDriver->pfnIdentify = PSPDataset::Identify;

            GetGDALDriverManager()->RegisterDriver( poDriver );
        }
    }

    PSPDataset::PSPDataset()
    {
        poImageDS = NULL;
        adfGeoTransform[0] = 0.0;
        adfGeoTransform[1] = 1.0;
        adfGeoTransform[2] = 0.0;
        adfGeoTransform[3] = 0.0;
        adfGeoTransform[4] = 0.0;
        adfGeoTransform[5] = 1.0;
    }

    PSPDataset::~PSPDataset()
    {
        if (poImageDS != NULL)
        {
            GDALClose( poImageDS );
            poImageDS = NULL;
        }
    }

    /************************************************************************/
    /*                          GetProjectionRef()                          */
    /************************************************************************/

    const char* PSPDataset::GetProjectionRef()
    {
        return "PROJCS[\"unnamed\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo_1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[23.92,-141.27,-80.9,-0,0.35,0.82,-0.12],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199433,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse_Mercator\"],PARAMETER[\"latitude_of_origin\",0],PARAMETER[\"central_meridian\",45],PARAMETER[\"scale_factor\",1],PARAMETER[\"false_easting\",8500000],PARAMETER[\"false_northing\",0],UNIT[\"Meter\",1]]";
    }


    CPLErr PSPDataset::GetGeoTransform( double * padfTransform )
    {
        memcpy(padfTransform, adfGeoTransform, 6 * sizeof(double));

        return CE_None;
    }

    void cp2utf( char* str, char* res ) {
        static const long utf[ 256 ] = {
            0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,
            31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,
            59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,
            87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,
            111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,1026,1027,8218,
            1107,8222,8230,8224,8225,8364,8240,1033,8249,1034,1036,1035,1039,1106,8216,8217,
            8220,8221,8226,8211,8212,8250,8482,1113,8250,1114,1116,1115,1119,160,1038,1118,1032,
            164,1168,166,167,1025,169,1028,171,172,173,174,1031,176,177,1030,1110,1169,181,182,
            183,1105,8470,1108,187,1112,1029,1109,1111,1040,1041,1042,1043,1044,1045,1046,1047,
            1048,1049,1050,1051,1052,1053,1054,1055,1056,1057,1058,1059,1060,1061,1062,1063,
            1064,1065,1066,1067,1068,1069,1070,1071,1072,1073,1074,1075,1076,1077,1078,1079,
            1080,1081,1082,1083,1084,1085,1086,1087,1088,1089,1090,1091,1092,1093,1094,1095,
            1096,1097,1098,1099,1100,1101,1102,1103
        };

        int cnt = strlen( str ),
        i = 0, j = 0;

        for(; i < cnt; ++i ) {
            long c = utf[ (unsigned char) str[ i ] ];
            if( c < 0x80 ) {
                res[ j++ ] = c;
            }
            else if( c < 0x800 ) {
                res[ j++ ] = c >> 6 | 0xc0;
                res[ j++ ] = c & 0x3f | 0x80;
            }
            else if( c < 0x10000 ) {
                res[ j++ ] = c >> 12 | 0xe0;
                res[ j++ ] = c >> 6 & 0x3f | 0x80;
                res[ j++ ] = c & 0x3f | 0x80;
            }
            else if( c < 0x200000 ) {
                res[ j++ ] = c >> 18 | 0xf0;
                res[ j++ ] = c >> 12 & 0x3f | 0x80;
                res[ j++ ] = c >> 6 & 0x3f | 0x80;
                res[ j++ ] = c & 0x3f | 0x80;
            }
        }
        res[ j ] = '\0';
    }

